// (C) Copyright 2008 Donald A. Graft, All Rights Reserved
//
// CUVIDServer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//  
// CUVIDServer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License
// along with GNU Make; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 

#define NVGLOBAL
#define VERSION "2.0.0 beta 1"

#include "windows.h"
#include "stdio.h"
#include "resource.h"
#include "CUVIDServer.h"

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

extern int decoder_open();
extern int decoder_close();
extern int decoder_reset();
extern int decoder_flush();
extern int decoder_set_mode(int deinterlace, int use_top_field, int use_pf, int codec_type,
							int resize_w, int resize_h, int crop_l, int crop_r, int crop_t, int crop_b);
extern int decoder_decode_nalu(unsigned char *buf, int len);

HANDLE hKill;

DWORD WINAPI ServerThread(LPVOID lpParam) 
{
	unsigned int *m = (unsigned int *)(&pMsg->rpc_message[0]);
	HANDLE arr[] = {hExec, hStopDecoding, hKill};
	HRESULT ci;
	DWORD stat;

	ci = CoInitialize(NULL);

	decoder_open();

	while (1)
	{
		stat = ::WaitForMultipleObjects(3, arr, FALSE, INFINITE);
		if (stat == 1)
		{
			decoder_reset();
			::ResetEvent(hStopDecoding);
			::ResetEvent(hFrameAvailable);
			continue;
		}
		else if (stat == 2)
			break;

		switch(m[0])
		{
		case OP_FLUSH:
			pMsg->rpc_result = decoder_flush();
			break;
		case OP_SET_MODE:
			pMsg->rpc_result = decoder_set_mode(m[1], m[2], m[3], m[4], m[5], m[6], m[7], m[8], m[9], m[10]);
			break;
		case OP_DECODE_NALU:
			pMsg->rpc_result = decoder_decode_nalu((unsigned char *) &m[2], m[1]);
			break;
		}
		::SetEvent(hDone);
	}

	decoder_reset();
	decoder_close();

	if (SUCCEEDED(ci))
		CoUninitialize();
	
	::ResetEvent(hKill);

    ExitThread(0); 
} 

HANDLE hLock;
HWND hwnd;
HINSTANCE		hInst;	// current instance
NOTIFYICONDATA	niData;	// notify icon data
#define TRAYICONID	1 //				ID number for the Notify Icon
#define SWM_TRAYMSG	WM_APP //		the message ID sent to our window
#define SWM_SHOW	WM_APP + 1 //	show the window
#define SWM_HIDE	WM_APP + 2 //	hide the window
#define SWM_EXIT	WM_APP + 3 //	close the window

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG  msg;    
	WNDCLASS wc = {0};
	HICON hIcon;
	HANDLE hServerThread;
	char ExePath[255];
	FILE *INIFile;
	char *ptr;

#if 0
	// Check for availability of GPU.
	hLock = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0,
						   4, "NVGlobalLock");
	if (hLock != NULL && GetLastError() == ERROR_ALREADY_EXISTS)
	{
		MessageBox(hwnd, "GPU is not available! Exiting...", NULL, MB_OK | MB_ICONERROR);
		exit(0);
	}
#endif

	GetModuleFileName(NULL, ExePath, 255);
	if ((ptr = strrchr(ExePath,'\\')) != 0) ptr++;
    else ptr = ExePath;
	*ptr = 0;
	strcat_s(ExePath, 255, "CUVIDServer.ini");
	if ((INIFile = fopen(ExePath, "r")) == NULL)
	{
		use_D3D = false;
	}
	else
	{
		fscanf(INIFile, "UseD3D=%d\n", &use_D3D);
	}

	hInst = hInstance;

	wc.lpszClassName = "Application";
	wc.hInstance     = hInstance;
	wc.hbrBackground = GetSysColorBrush(COLOR_3DFACE);
	wc.lpfnWndProc   = WndProc;
	RegisterClass(&wc);

	hwnd = CreateWindow(wc.lpszClassName, "CUVID Server " VERSION,
						WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX,
						100, 100, 400, 100, 0, 0, hInstance, 0);  
	hIcon = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ICON)); 
	SendMessage(hwnd, (UINT) WM_SETICON, (WPARAM) ICON_SMALL, (LPARAM) hIcon); 
	SendMessage(hwnd, (UINT) WM_SETICON, (WPARAM) ICON_BIG, (LPARAM) hIcon);

	hMap = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0,
							   sizeof(TSharedMemory),
							   "NVSharedMem");
	if (hMap == NULL)
	{
		DestroyWindow(hwnd);
		return 0;
	}
	if (GetLastError() == ERROR_ALREADY_EXISTS)
	{
		CloseHandle(hMap);
		DestroyWindow(hwnd);
		return 0;
	}
	pMsg = (TSharedMemory*) ::MapViewOfFile(hMap, FILE_MAP_WRITE, 0, 0, sizeof(TSharedMemory));
	if (pMsg == NULL)
	{
		CloseHandle(hMap);
		DestroyWindow(hwnd);
		return 0;
	}

	// Place the application minimized in the system tray.
	niData.cbSize = sizeof(NOTIFYICONDATA);
	niData.uID = TRAYICONID;
	niData.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
	niData.hIcon = (HICON) LoadImage(hInstance,MAKEINTRESOURCE(IDD_ICON),
					IMAGE_ICON, GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), LR_DEFAULTCOLOR);
	niData.hWnd = hwnd;
    niData.uCallbackMessage = SWM_TRAYMSG;
    lstrcpyn(niData.szTip, "CUVID Server " VERSION, sizeof(niData.szTip)/sizeof(TCHAR));
	Shell_NotifyIcon(NIM_ADD,&niData);
	if(niData.hIcon && DestroyIcon(niData.hIcon))
		niData.hIcon = NULL;

	// Create events and server thread.
	hExec = ::CreateEvent(NULL, FALSE, FALSE, "NVExec");
	hDone = ::CreateEvent(NULL, FALSE, FALSE, "NVDone");
	hStopDecoding = ::CreateEvent(NULL, TRUE, FALSE, "NVStopDecoding");
	hFrameRequested = ::CreateEvent(NULL, FALSE, FALSE, "NVRequested");
	hFrameAvailable = ::CreateEvent(NULL, FALSE, FALSE, "NVAvailable");
	hKill = ::CreateEvent(NULL, TRUE, FALSE, 0);
	hServerThread = CreateThread(NULL, 0, ServerThread, NULL, 0, NULL);

	while (GetMessage(&msg, NULL, 0, 0))
		DispatchMessage(&msg);

	::SetEvent(hKill);
	::SetEvent(hStopDecoding);
	while (::WaitForSingleObject(hKill, 0) == WAIT_OBJECT_0)
		Sleep(10);

	niData.uFlags = 0;
	Shell_NotifyIcon(NIM_DELETE,&niData);

	CloseHandle(hLock);
	UnmapViewOfFile(pMsg);
	CloseHandle(hMap);
	CloseHandle(hExec);
	CloseHandle(hDone);
	CloseHandle(hStopDecoding);
	CloseHandle(hFrameRequested);
	CloseHandle(hFrameAvailable);
	CloseHandle(hServerThread);
	CloseHandle(hKill);

	return (int) msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	HMENU hMenu;
	HICON hIcon;
	POINT pt;

	switch(msg)  
	{
	case SWM_TRAYMSG:
		switch(lParam)
		{
		case WM_LBUTTONDBLCLK:
			ShowWindow(hwnd, SW_RESTORE);
			break;
		case WM_RBUTTONDOWN:
		case WM_CONTEXTMENU:
			GetCursorPos(&pt);
			hMenu = CreatePopupMenu();
			if(hMenu)
			{
				if(IsWindowVisible(hwnd))
					InsertMenu(hMenu, -1, MF_BYPOSITION, SWM_HIDE, "Hide");
				else
					InsertMenu(hMenu, -1, MF_BYPOSITION, SWM_SHOW, "Show");
				InsertMenu(hMenu, -1, MF_BYPOSITION, SWM_EXIT, "Exit");
				SetForegroundWindow(hwnd);
				TrackPopupMenu(hMenu, TPM_BOTTOMALIGN, pt.x, pt.y, 0, hwnd, NULL);
				DestroyMenu(hMenu);
			}
		}
		break;

	case WM_SYSCOMMAND:
		if((wParam & 0xFFF0) == SC_MINIMIZE)
		{
			ShowWindow(hwnd, SW_HIDE);
			return 1;
		}

	case WM_COMMAND:
		wmId    = LOWORD(wParam);
		wmEvent = HIWORD(wParam); 

		switch (wmId)
		{
		case SWM_SHOW:
			ShowWindow(hwnd, SW_RESTORE);
			break;
		case SWM_HIDE:
		case IDOK:
			ShowWindow(hwnd, SW_HIDE);
			break;
		case SWM_EXIT:
			PostQuitMessage(0);
			break;
		}
		break;

	case WM_INITDIALOG:
		hIcon = (HICON)LoadImage(hInst,
			MAKEINTRESOURCE(IDD_ICON),
			IMAGE_ICON, 0,0, LR_SHARED|LR_DEFAULTSIZE);
		SendMessage(hwnd,WM_SETICON,ICON_BIG,(LPARAM)hIcon);
		SendMessage(hwnd,WM_SETICON,ICON_SMALL,(LPARAM)hIcon);
		break;

	case WM_CREATE:
		CreateWindow("static", "CUVID Server remains alive as long\nas this application is running.",
			WS_CHILD | WS_VISIBLE | SS_CENTER,
			0, 16, 370, 100, hwnd, NULL, NULL, NULL);
		break;

	case WM_DESTROY:
		PostQuitMessage(0);
		break; 
	}
	return DefWindowProc(hwnd, msg, wParam, lParam);
}
