// (C) Copyright 2008 Donald A. Graft, All Rights Reserved
//
// CUVIDServer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//  
// CUVIDServer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License
// along with GNU Make; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 

// Video decoder for Nvidia GPUs. Portions provided by Nvidia Corporation.
// The copyright notice above applies to the parts added by DAG.

#include "windows.h"
#include "stdio.h"
#include "d3d9.h"
#include "nvcuvid.h"
#include "cudad3d9.h"
#include "CUVIDServer.h"

#define MAX_FRM_CNT     16
#define DISPLAY_DELAY	2

typedef struct
{
	cudaVideoCodec cuCodec;
	CUvideoparser cuParser;
	CUvideodecoder cuDecoder;
	CUstream cuStream;
	CUVIDDECODECREATEINFO dci;
	CUVIDPARSERDISPINFO DisplayQueue[DISPLAY_DELAY];
	int display_pos;
	int deinterlace;
	int use_top_field;
	int use_pf;
	int resize_w, resize_h;
	int crop_l, crop_r, crop_t, crop_b;
} DecodeSession;

DecodeSession Session;

IDirect3D9 *g_pD3D;
IDirect3DDevice9 *g_pD3Dev;
CUcontext g_cuContext;
CUdevice g_cuDevice;
CUVIDPARSERPARAMS parserInitParams;

int ForceSequenceUpdate;
int MyPOCs[MAX_FRM_CNT];
int MyIntraFrameFlag[MAX_FRM_CNT];
int MyMbaffFlag[MAX_FRM_CNT];
int DisplayPicture(DecodeSession *state, CUVIDPARSERDISPINFO *pPicParams);
extern HANDLE hFrameRequested;
extern HANDLE hFrameAvailable;
extern HWND hwnd;

//////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Parser callbacks
//

// Called when the decoder encounters a video format change (or initial sequence header)
int CUDAAPI HandleVideoSequence(void *pvUserData, CUVIDEOFORMAT *pFormat)
{
    DecodeSession *state = (DecodeSession *)pvUserData;
	CUresult result;
	bool crop, resize;

    if ((pFormat->codec != state->dci.CodecType)
		|| (pFormat->coded_width != state->dci.ulWidth)
		|| (pFormat->coded_height != state->dci.ulHeight)
		|| (pFormat->chroma_format != state->dci.ChromaFormat)
		|| ForceSequenceUpdate)
    {
		ForceSequenceUpdate = 0;
		if (state->cuDecoder)
		{
            cuvidDestroyDecoder(state->cuDecoder);
            state->cuDecoder = NULL;
		}
		memset(&state->dci, 0, sizeof(CUVIDDECODECREATEINFO));
		state->dci.ulWidth = pFormat->coded_width;
		state->dci.ulHeight = pFormat->coded_height;
		state->dci.ulNumDecodeSurfaces = MAX_FRM_CNT;
		state->dci.CodecType = pFormat->codec;
		state->dci.ChromaFormat = pFormat->chroma_format;
		// Output (pass through)
		state->dci.OutputFormat = cudaVideoSurfaceFormat_NV12;
		if (Session.deinterlace != WEAVE)
			state->dci.DeinterlaceMode = cudaVideoDeinterlaceMode_Adaptive;
		else
			state->dci.DeinterlaceMode = cudaVideoDeinterlaceMode_Weave;
		state->dci.ulNumOutputSurfaces = 1;

		// Cropping and resizing.
		crop = resize = false;
		if (Session.crop_l || Session.crop_r || Session.crop_t || Session.crop_b)
			crop = true;
		if (Session.resize_w && Session.resize_h)
			resize = true;
		if (crop == false && resize == false)
		{
			state->dci.ulTargetWidth = pFormat->coded_width;
			state->dci.ulTargetHeight = pFormat->coded_height;
			state->dci.display_area.left = 0;
			state->dci.display_area.right = pFormat->coded_width;
			state->dci.display_area.top = 0;
			state->dci.display_area.bottom = pFormat->coded_height;
		}
		else if (crop == false && resize == true)
		{
			state->dci.ulTargetWidth = Session.resize_w;
			state->dci.ulTargetHeight = Session.resize_h;
			state->dci.display_area.left = 0;
			state->dci.display_area.right = pFormat->coded_width;
			state->dci.display_area.top = 0;
			state->dci.display_area.bottom = pFormat->coded_height;
		}
		else if (crop == true && resize == false)
		{
			state->dci.ulTargetWidth = pFormat->coded_width - Session.crop_r - Session.crop_l;
			state->dci.ulTargetHeight = pFormat->coded_height - Session.crop_b - Session.crop_t;
			state->dci.display_area.left = Session.crop_l;
			state->dci.display_area.right = pFormat->coded_width - Session.crop_r;
			state->dci.display_area.top = Session.crop_t;
			state->dci.display_area.bottom = pFormat->coded_height - Session.crop_b;
		}
		else if (crop == true && resize == true)
		{
			state->dci.ulTargetWidth = Session.resize_w;
			state->dci.ulTargetHeight = Session.resize_h;
			state->dci.display_area.left = Session.crop_l;
			state->dci.display_area.right = pFormat->coded_width - Session.crop_r;
			state->dci.display_area.top = Session.crop_t;
			state->dci.display_area.bottom = pFormat->coded_height - Session.crop_b;
		}

		memset(state->dci.Reserved1, 0, sizeof(state->dci.Reserved1));
		memset(state->dci.Reserved2, 0, sizeof(state->dci.Reserved2));
#if 0
		{
			char buf[80];
			sprintf(buf,"CUVIDServer: ulWidth = %d", state->dci.ulWidth);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: ulHeight = %d", state->dci.ulHeight);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: ulNumDecodeSurfaces = %d", state->dci.ulNumDecodeSurfaces);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: CodecType = %d", state->dci.CodecType);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: ChromaFormat = %d", state->dci.ChromaFormat);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: ulCreationFlags = %d", state->dci.ulCreationFlags);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: display_area.left = %d", state->dci.display_area.left);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: display_area.right = %d", state->dci.display_area.right);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: display_area.top = %d", state->dci.display_area.top);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: display_area.bottom = %d", state->dci.display_area.bottom);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: OutputFormat = %d", state->dci.OutputFormat);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: DeinterlaceMode = %d", state->dci.DeinterlaceMode);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: ulTargetWidth = %d", state->dci.ulTargetWidth);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: ulTargetHeight = %d", state->dci.ulTargetHeight);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: ulNumOutputSurfaces = %d", state->dci.ulNumOutputSurfaces);
			OutputDebugString(buf);
			sprintf(buf,"CUVIDServer: vidLock = %d", state->dci.vidLock);
			OutputDebugString(buf);
		}
#endif

		// Create the decoder
		result = cuvidCreateDecoder(&state->cuDecoder, &state->dci);
		if (CUDA_SUCCESS != result)
		{
			char buf[80];
			sprintf_s(buf, 80, "Failed to create video decoder (%d).", result);
			MessageBox(hwnd, buf, "Server", MB_OK | MB_ICONERROR | MB_TOPMOST | MB_SETFOREGROUND);
			exit(0);
		}
	}
	return 1;
}

// Called by the video parser to decode a single picture
// Since the parser will deliver data as fast as it can, we need to make sure that the picture
// index we're attempting to use for decode is no longer used for display
int CUDAAPI HandlePictureDecode(void *pvUserData, CUVIDPICPARAMS *pPicParams)
{
    DecodeSession *state = (DecodeSession *)pvUserData;
    CUresult result;
	int flush_pos;

    if (pPicParams->CurrPicIdx < 0) // Should never happen
    {
		MessageBox(hwnd, "Invalid picture index.\n", "Server", MB_OK | MB_ICONERROR | MB_TOPMOST | MB_SETFOREGROUND);
		exit(0);
    }
    // Make sure that the new frame we're decoding into is not still in the display queue
    // (this could happen if we do not have enough free frame buffers to handle the max delay)
    flush_pos = state->display_pos; // oldest frame
    for (;;)
    {
        bool frame_in_use = false;
        for (int i=0; i<DISPLAY_DELAY; i++)
        {
            if (state->DisplayQueue[i].picture_index == pPicParams->CurrPicIdx)
            {
                frame_in_use = true;
                break;
            }
        }
        if (!frame_in_use)
        {
            // No problem: we're safe to use this frame
            break;
        }
        // The target frame is still pending in the display queue:
        // Flush the oldest entry from the display queue and repeat
        if (state->DisplayQueue[flush_pos].picture_index >= 0)
        {
            DisplayPicture(state, &state->DisplayQueue[flush_pos]);
            state->DisplayQueue[flush_pos].picture_index = -1;
        }
        flush_pos = (flush_pos + 1) % DISPLAY_DELAY;
    }
	if ((!pPicParams->field_pic_flag) || (!pPicParams->second_field))
		MyIntraFrameFlag[pPicParams->CurrPicIdx] = pPicParams->intra_pic_flag;
	MyPOCs[pPicParams->CurrPicIdx] = pPicParams->CodecSpecific.h264.CurrFieldOrderCnt[0] |
										pPicParams->CodecSpecific.h264.CurrFieldOrderCnt[1];
	MyMbaffFlag[pPicParams->CurrPicIdx] = pPicParams->CodecSpecific.h264.MbaffFrameFlag;
	result = cuvidDecodePicture(state->cuDecoder, pPicParams);
    if (result != CUDA_SUCCESS)
    {
		MessageBox(hwnd, "cuvidDecodePicture() error.", "Server", MB_OK | MB_ICONEXCLAMATION | MB_TOPMOST | MB_SETFOREGROUND );
		// Satisfy the frame request, because this decode won't result in a HandlePictureDisplay() callback.
		// This prevents the client from locking up.
		::SetEvent(hFrameAvailable);
	}
    return (result == CUDA_SUCCESS);
}

// Called by the video parser to display a video frame (in the case of field pictures, there may be
// 2 decode calls per 1 display call, since two fields make up one frame)
int CUDAAPI HandlePictureDisplay(void *pvUserData, CUVIDPARSERDISPINFO *pPicParams)
{
    DecodeSession *state = (DecodeSession *)pvUserData;

    if (state->DisplayQueue[state->display_pos].picture_index >= 0)
    {
        DisplayPicture(state, &state->DisplayQueue[state->display_pos]);
        state->DisplayQueue[state->display_pos].picture_index = -1;
    }
    state->DisplayQueue[state->display_pos] = *pPicParams;
    state->display_pos = (state->display_pos + 1) % DISPLAY_DELAY;
    return TRUE;
}

int DisplayPicture(DecodeSession *state, CUVIDPARSERDISPINFO *pPicParams)
{
    CUVIDPROCPARAMS vpp;
    CUdeviceptr devPtr;
    CUresult result;
    unsigned int pitch = 0, nvw, nvh;
	int nv12_size;
	HANDLE arr[] = {hFrameRequested, hStopDecoding};

	if (Session.use_pf == false && Session.deinterlace > 0)
		vpp.progressive_frame = false;
	else
		vpp.progressive_frame = pPicParams->progressive_frame;
    vpp.top_field_first = pPicParams->top_field_first;
	if (Session.deinterlace == DEINTERLACE)
	{
		if ((Session.use_top_field && pPicParams->top_field_first) ||
			(!Session.use_top_field && !pPicParams->top_field_first))
			vpp.second_field = false;
		else
			vpp.second_field = true;
	}
	else
	{
		vpp.second_field = false;
	}
	vpp.unpaired_field = false;
    vpp.reserved_flags = 0;
    vpp.raw_input_dptr = 0;
    vpp.raw_input_pitch = 0;
	memset(vpp.Reserved, 0, sizeof(vpp.Reserved));
	memset(vpp.Reserved3, 0, sizeof(vpp.Reserved3));

	::WaitForMultipleObjects(2, arr, FALSE, INFINITE);
    result = cuvidMapVideoFrame(state->cuDecoder, pPicParams->picture_index, &devPtr, &pitch, &vpp);
    if (result != CUDA_SUCCESS)
    {
		MessageBox(hwnd, "cuvidMapVideoFrame() error.", "Server", MB_OK | MB_ICONERROR | MB_TOPMOST | MB_SETFOREGROUND);
        return 0;
    }
    nvw = state->dci.ulTargetWidth;
    nvh = state->dci.ulTargetHeight;
    nv12_size = pitch * (nvh + nvh/2); // 12bpp
	pMsg->width = nvw;
	pMsg->height = nvh;
	pMsg->pitch = pitch;
	pMsg->progressive = pPicParams->progressive_frame;
	cuMemcpyDtoH(pMsg->nv12_frame, devPtr, nv12_size);
    cuvidUnmapVideoFrame(state->cuDecoder, devPtr);
	::SetEvent(hFrameAvailable);

	if (Session.deinterlace != BOB)
		return 1;

	// Bobbing.
	vpp.second_field = true;
	if (::WaitForMultipleObjects(2, arr, FALSE, INFINITE) == 1)
		return 1;
    result = cuvidMapVideoFrame(state->cuDecoder, pPicParams->picture_index, &devPtr, &pitch, &vpp);
    if (result != CUDA_SUCCESS)
    {
		MessageBox(hwnd, "cuvidMapVideoFrame() error.", "Server", MB_OK | MB_ICONERROR | MB_TOPMOST | MB_SETFOREGROUND);
        return 0;
    }
	cuMemcpyDtoH(pMsg->nv12_frame, devPtr, nv12_size);
    cuvidUnmapVideoFrame(state->cuDecoder, devPtr);
	::SetEvent(hFrameAvailable);
    return 1;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// CUDA 2.0 initialization
//

bool InitCuda()
{
    CUresult err;
	int i;
	int lAdapter, lAdapterCount;

    err = cuInit(0);
    if (err != CUDA_SUCCESS)
    {
		MessageBox(hwnd, "cuInit() failed.", "Server", MB_OK | MB_ICONERROR | MB_TOPMOST | MB_SETFOREGROUND);
        return false;
    }

	if (use_D3D == true)
	{
		D3DPRESENT_PARAMETERS d3dpp;
		HRESULT hr;
		HWND hWnd;

		// Create an instance of Direct3D.
		g_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
		if (g_pD3D == NULL)
		{
			MessageBox(hwnd, "Failed to create D3D9.", "Server", MB_OK | MB_ICONERROR | MB_TOPMOST | MB_SETFOREGROUND);
			return false;
		}

		lAdapterCount = g_pD3D->GetAdapterCount();
		for (lAdapter=0; lAdapter<lAdapterCount; lAdapter++)
		{
			// Create the Direct3D9 device and the swap chain. In this example, the swap 
			// chain is the same size as the current display mode. The format is RGB-32.
			ZeroMemory(&d3dpp, sizeof(d3dpp));
			d3dpp.Windowed = TRUE;
			d3dpp.BackBufferFormat = D3DFMT_X8R8G8B8;
			d3dpp.BackBufferWidth = 640;
			d3dpp.BackBufferHeight = 480;
			d3dpp.BackBufferCount = 1;
			d3dpp.SwapEffect = D3DSWAPEFFECT_COPY;
			d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
			d3dpp.Flags = D3DPRESENTFLAG_VIDEO;//D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
			hWnd = GetDesktopWindow();
			hr = g_pD3D->CreateDevice(lAdapter,
									  D3DDEVTYPE_HAL,
									  hWnd,
									  D3DCREATE_FPU_PRESERVE | D3DCREATE_MULTITHREADED | D3DCREATE_HARDWARE_VERTEXPROCESSING,
									  &d3dpp,
									  &g_pD3Dev);
			if (hr == S_OK)
			{
				err = cuD3D9CtxCreate(&g_cuContext, &g_cuDevice, 0, g_pD3Dev);
				if (err == CUDA_SUCCESS)
				{
					// Init display queue
					for (i=0; i<DISPLAY_DELAY; i++)
					{
						Session.DisplayQueue[i].picture_index = -1;   // invalid
					}
					Session.display_pos = 0;
					return true;
				}                
				g_pD3Dev->Release();
				g_pD3Dev = NULL;
			}
		}
		MessageBox(hwnd, "Failed to find CUDA-compatible D3D device.", "Server", MB_OK | MB_ICONERROR | MB_TOPMOST | MB_SETFOREGROUND);
		return false;
	}
	else
	{
		cuDeviceGetCount(&lAdapterCount);
		for (lAdapter = 0; lAdapter < lAdapterCount; lAdapter++)
		{
			err = cuDeviceGet(&g_cuDevice, lAdapter);
			if (err != CUDA_SUCCESS)
				continue;
			err = cuCtxCreate(&g_cuContext, 0, g_cuDevice);
			if (err == CUDA_SUCCESS)
			{
				// Init display queue
				for (i=0; i<DISPLAY_DELAY; i++)
				{
					Session.DisplayQueue[i].picture_index = -1;   // invalid
				}
				Session.display_pos = 0;

				return true;
			}
		}
		MessageBox(hwnd, "cuCtxCreate() failed.", "Server", MB_OK | MB_ICONERROR | MB_TOPMOST | MB_SETFOREGROUND);
		return false;
	}
}

bool DeinitCuda()
{
    if (g_cuContext)
    {
        cuCtxDestroy(g_cuContext);
        g_cuContext = NULL;
    }
	if (use_D3D == true)
	{
		if (g_pD3Dev)
		{
			g_pD3Dev->Release();
			g_pD3Dev = NULL;
		}
		if (g_pD3D)
		{
			g_pD3D->Release();
			g_pD3D = NULL;
		}
	}
    return true;
}

// ------------- Implementation of public interface ----------------------------------

int decoder_close(void)
{
	DeinitCuda();

    // Delete all created objects
    if (Session.cuParser != NULL)
    {
        cuvidDestroyVideoParser(Session.cuParser);
        Session.cuParser = NULL;
    }
    if (Session.cuDecoder != NULL)
    {
        cuvidDestroyDecoder(Session.cuDecoder);
        Session.cuDecoder = NULL;
    }
    if (Session.cuStream != NULL)
    {
        cuStreamDestroy(Session.cuStream);
        Session.cuStream = NULL;
    }
	return 0;
}

int decoder_flush(void)
{
    CUVIDSOURCEDATAPACKET pkt;

	// Nothing to do if CUDA hasn't been started yet.
	if (!Session.cuParser)
		return 0;

	pkt.flags = CUVID_PKT_ENDOFSTREAM;
    pkt.payload_size = 0;
    pkt.payload = NULL;
    pkt.timestamp = 0;  // not using timestamps
    cuvidParseVideoData(Session.cuParser, &pkt);

	// Flush display queue
    for (int i=0; i<DISPLAY_DELAY; i++)
    {
        if (Session.DisplayQueue[Session.display_pos].picture_index >= 0)
        {
            DisplayPicture(&Session, &Session.DisplayQueue[Session.display_pos]);
            Session.DisplayQueue[Session.display_pos].picture_index = -1;
        }
        Session.display_pos = (Session.display_pos + 1) % DISPLAY_DELAY;
    }

	return 0;
}

int decoder_reset(void)
{
    CUresult result;
	int i;

	if (Session.cuParser != NULL)
	{
		cuvidDestroyVideoParser(Session.cuParser);
		Session.cuParser = NULL;
	}
	// Create video parser
	memset(&parserInitParams, 0, sizeof(parserInitParams));
	parserInitParams.CodecType = Session.cuCodec;
	parserInitParams.ulMaxNumDecodeSurfaces = MAX_FRM_CNT;
	parserInitParams.ulErrorThreshold = 100;
	parserInitParams.ulMaxDisplayDelay = 4;
	parserInitParams.pUserData = &Session;
	parserInitParams.pfnSequenceCallback = HandleVideoSequence;
	parserInitParams.pfnDecodePicture = HandlePictureDecode;
	parserInitParams.pfnDisplayPicture = HandlePictureDisplay;
	result = cuvidCreateVideoParser(&Session.cuParser, &parserInitParams);
	if (result != CUDA_SUCCESS)
	{
		decoder_close();
		return -1;
	}
	// Flush display queue
    for (i=0; i<DISPLAY_DELAY; i++)
        Session.DisplayQueue[i].picture_index = -1;
    Session.display_pos = 0;

	return 0;
}

int decoder_decode_nalu(unsigned char* buf, int len)
{
    CUVIDSOURCEDATAPACKET pkt;

	pkt.flags = 0;
    pkt.payload_size = (unsigned long) len;
    pkt.payload = buf;
    pkt.timestamp = 0;  // not using timestamps
    cuvidParseVideoData(Session.cuParser, &pkt);

	return len;
}

int decoder_set_mode(int deinterlace, int use_top_field, int use_pf, int codec_type,
					 int resize_w, int resize_h, int crop_l, int crop_r, int crop_t, int crop_b)
{
	Session.deinterlace = deinterlace;
	Session.use_top_field = use_top_field;
	Session.use_pf = use_pf;
	switch (codec_type)
	{
	case 0:
		Session.cuCodec = cudaVideoCodec_H264;
		break;
	case 1:
		Session.cuCodec = cudaVideoCodec_VC1;
		break;
	case 2:
		Session.cuCodec = cudaVideoCodec_MPEG1;
		break;
	case 3:
		Session.cuCodec = cudaVideoCodec_MPEG2;
		break;
	}
	Session.resize_w = resize_w;
	Session.resize_h = resize_h;
	Session.crop_l = crop_l;
	Session.crop_r = crop_r;
	Session.crop_t = crop_t;
	Session.crop_b = crop_b;
	ForceSequenceUpdate = 1;
	return 0;
}

int decoder_open()
{
	int i;
    CUresult result;

	memset(&Session, 0, sizeof(Session));

	// Start assuming MPEG2. It can be changed with decoder_set_mode().
	Session.cuCodec = cudaVideoCodec_MPEG2;
    
    // Initialize CUDA 2.0.
    if (!InitCuda())
    {
		decoder_close();
		return -1;
    }
	// Create video parser
    memset(&parserInitParams, 0, sizeof(parserInitParams));
    parserInitParams.CodecType = Session.cuCodec;
    parserInitParams.ulMaxNumDecodeSurfaces = MAX_FRM_CNT;
	parserInitParams.ulErrorThreshold = 100;
	parserInitParams.ulMaxDisplayDelay = 4;
	parserInitParams.pUserData = &Session;
    parserInitParams.pfnSequenceCallback = HandleVideoSequence;
    parserInitParams.pfnDecodePicture = HandlePictureDecode;
    parserInitParams.pfnDisplayPicture = HandlePictureDisplay;
    result = cuvidCreateVideoParser(&Session.cuParser, &parserInitParams);
    if (result != CUDA_SUCCESS)
    {
		decoder_close();
		return -1;
    }
    result = cuStreamCreate(&Session.cuStream, 0);
    if (result != CUDA_SUCCESS)
    {
		decoder_close();
        return -1;
    }
    for (i=0; i<DISPLAY_DELAY; i++)
        Session.DisplayQueue[i].picture_index = -1;

	return 0;
}
